#include "precomp.h"

CCP210xDevices::CCP210xDevices(HRESULT* phr)
    : m_pTypeInfo(nullptr), m_lRefCount(1)
{
    ::LockServer();

    ITypeLib* pTlb = nullptr;
    if(SUCCEEDED((*phr = LoadRegTypeLib(LIBID_CP210xExLib, 1, 0, LOCALE_USER_DEFAULT, &pTlb))))
    {
        *phr = pTlb->GetTypeInfoOfGuid(IID_ICP210xDevices, &m_pTypeInfo);
        pTlb->Release();
    }

    DWORD dw;
    ::CP210x_GetNumDevices(&dw);
}

CCP210xDevices::~CCP210xDevices()
{
    if(m_pTypeInfo)
        m_pTypeInfo->Release();

    ::UnlockServer();
}

IFACEMETHODIMP
CCP210xDevices::QueryInterface(REFIID riid, void** ppvObject)
{
    __try
    {
        *ppvObject = nullptr;

        if(IsEqualIID(riid, IID_IUnknown))
            *ppvObject = GetUnknown();
        else if(IsEqualIID(riid, IID_IDispatch))
            *ppvObject = static_cast<IDispatch*>(this);
        else if(IsEqualIID(riid, IID_ICP210xDevices))
            *ppvObject = static_cast<ICP210xDevices*>(this);
        else if(IsEqualIID(riid, IID_IProvideClassInfo))
            *ppvObject = static_cast<IProvideClassInfo*>(this);
        else if(IsEqualIID(riid, IID_ISupportErrorInfo))
            *ppvObject = static_cast<ISupportErrorInfo*>(this);

        if(*ppvObject)
        {
            static_cast<IUnknown*>(*ppvObject)->AddRef();
            return(S_OK);
        }
    }
    __except((GetExceptionCode() == EXCEPTION_ACCESS_VIOLATION) ? EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH)
    {
        return(E_INVALIDARG);
    }

    return(E_NOINTERFACE);
}

IFACEMETHODIMP_(ULONG)
CCP210xDevices::AddRef()
{
    return(InterlockedIncrement(&m_lRefCount));
}

IFACEMETHODIMP_(ULONG)
CCP210xDevices::Release()
{
    ULONG ulResult;

    if(0 == (ulResult = InterlockedDecrement(&m_lRefCount)))
        delete this;

    return(ulResult);
}

IFACEMETHODIMP
CCP210xDevices::GetTypeInfoCount(UINT* pctinfo)
{
    if(IsBadWritePtrT(pctinfo))
        return(E_POINTER);

    *pctinfo = 1;
    return(S_OK);
}

IFACEMETHODIMP
CCP210xDevices::GetTypeInfo(UINT iTInfo, LCID, ITypeInfo** ppTInfo)
{
    if(IsBadWritePtrT(ppTInfo))
        return(E_POINTER);

    if(iTInfo)
    {
        *ppTInfo = nullptr;
        return(DISP_E_BADINDEX);
    }

    *ppTInfo = m_pTypeInfo;
    (*ppTInfo)->AddRef();
    return(S_OK);
}

IFACEMETHODIMP
CCP210xDevices::GetIDsOfNames(REFIID, LPOLESTR* rgszNames, UINT cNames, LCID, DISPID* rgDispId)
{
    return(m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId));
}

IFACEMETHODIMP
CCP210xDevices::Invoke(DISPID dispIdMember, REFIID, LCID lcid, WORD wFlags, DISPPARAMS* pDispParams, VARIANT* pVarResult, EXCEPINFO* pExcepInfo, UINT* puArgErr)
{
    CThreadLocale loc(lcid);
    UINT uArgErr;
    return(m_pTypeInfo->Invoke(static_cast<ICP210xDevices*>(this), dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, (puArgErr) ? puArgErr : &uArgErr));
}

IFACEMETHODIMP
CCP210xDevices::get_Item(const VARIANT varIndex, VARIANT* pVal)
{
    CCP210x*    pDevice;
    VARIANT     var;
    HRESULT     hr;

    if(IsBadWritePtrT(pVal))
        return(RaiseCOMException(DISPID_CP210X_ITEM, E_POINTER));

    VariantInit(pVal);
    VariantInit(&var);

    if(FAILED((hr = VariantChangeType(&var, &varIndex, 0, VT_UI4))))
        return(RaiseCOMException(DISPID_CP210X_ITEM, hr));

    if(nullptr == (pDevice = new (std::nothrow) CCP210x(this, V_UI4(&var), &hr)))
        return(RaiseCOMException(DISPID_CP210X_ITEM, E_OUTOFMEMORY));

    if(SUCCEEDED(hr))
    {
        if(SUCCEEDED((hr = pDevice->QueryInterface(IID_IDispatch, (void**)&V_UNKNOWN(pVal)))))
            V_VT(pVal) = VT_DISPATCH;
    }

    pDevice->Release();
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_ITEM, hr));
}

IFACEMETHODIMP
CCP210xDevices::get_Count(long* pVal)
{
    return(getNumDevices(DISPID_CP210X_COUNT, pVal));
}

IFACEMETHODIMP
CCP210xDevices::get_length(long* pVal)
{
    return(getNumDevices(DISPID_CP210X_LENGTH, pVal));
}

HRESULT
CCP210xDevices::getNumDevices(DISPID dispid, long* pVal) noexcept
{
    if(IsBadWritePtrT(pVal))
        return(RaiseCOMException(dispid, E_POINTER));

    const CP210x_STATUS stat = ::CP210x_GetNumDevices(reinterpret_cast<DWORD*>(pVal));
    return((CP210x_SUCCESS == stat) ? S_OK : RaiseCOMException(dispid, ::StatusToHresult(stat)));
}

IFACEMETHODIMP
CCP210xDevices::get__NewEnum(IUnknown** ppVal)
{
    CEnumVARIANT*   pEnum;
    HRESULT         hr;

    if(IsBadWritePtrT(ppVal))
        return(RaiseCOMException(DISPID_CP210X__NEWENUM, E_POINTER));

    *ppVal = nullptr;

    if(nullptr == (pEnum = new (std::nothrow) CEnumVARIANT(this)))
        return(RaiseCOMException(DISPID_CP210X__NEWENUM, E_OUTOFMEMORY));

    hr = pEnum->QueryInterface(IID_IUnknown, reinterpret_cast<void**>(ppVal));
    pEnum->Release();
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X__NEWENUM, hr));
}

IFACEMETHODIMP
CCP210xDevices::GetProductString(long nDeviceNum, enum Options nOptions, BSTR* pbstrString)
{
    CP210x_DEVICE_STRING    DevString;
    CP210x_STATUS           stat;

    if(IsBadWritePtrT(pbstrString))
        return(RaiseCOMException(DISPID_CP210X_GETPRODUCTSTRING, E_POINTER));

    *pbstrString = nullptr;

    if(CP210x_SUCCESS != (stat = ::CP210x_GetProductString(static_cast<DWORD>(nDeviceNum), DevString, static_cast<DWORD>(nOptions))))
        return(RaiseCOMException(DISPID_CP210X_GETPRODUCTSTRING, ::StatusToHresult(stat)));

    const HRESULT hr = BSTRFromStringA(DevString, pbstrString);
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_GETPRODUCTSTRING, hr));
}

IFACEMETHODIMP
CCP210xDevices::GetClassInfo(ITypeInfo** ppTI)
{
    ITypeLib*   pTlb = nullptr;
    HRESULT     hr;
    UINT        uIndex;

    if(IsBadWritePtrT(ppTI))
        return(E_POINTER);

    *ppTI = nullptr;

    if(SUCCEEDED((hr = m_pTypeInfo->GetContainingTypeLib(&pTlb, &uIndex))))
    {
        hr = pTlb->GetTypeInfoOfGuid(CLSID_CP210xDevices, ppTI);
        pTlb->Release();
    }

    return(hr);
}

IFACEMETHODIMP
CCP210xDevices::InterfaceSupportsErrorInfo(REFIID riid)
{
    __try
    {
        if(IsEqualIID(riid, IID_ICP210xDevices))
            return(S_OK);
    }
    __except((GetExceptionCode() == EXCEPTION_ACCESS_VIOLATION) ? EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH)
    {
        return(E_INVALIDARG);
    }

    return(S_FALSE);
}
