#include "precomp.h"

#define ENTRY_FILENAME  MAKEINTRESOURCE(1)

static const LPCTSTR RegEntryTable[][3] =
{
    { TEXT("CLSID\\{2FC05A0F-CD36-4A77-A788-C9BC232E947F}"),                    nullptr,                TEXT("CP210x")                                  },
    { TEXT("CLSID\\{2FC05A0F-CD36-4A77-A788-C9BC232E947F}\\TypeLib"),           nullptr,                TEXT("{0A49C6AF-546F-4381-893E-28CD124966C2}")  },
    { TEXT("CLSID\\{2FC05A0F-CD36-4A77-A788-C9BC232E947F}\\ProgID"),            nullptr,                TEXT("CP210x.CP210xDevices")                    },
    { TEXT("CLSID\\{2FC05A0F-CD36-4A77-A788-C9BC232E947F}\\InprocServer32"),    nullptr,                ENTRY_FILENAME                                  },
    { TEXT("CLSID\\{2FC05A0F-CD36-4A77-A788-C9BC232E947F}\\InprocServer32"),    TEXT("ThreadingModel"), TEXT("Both")                                    },
    { TEXT("CP210x.CP210xDevices"),                                             nullptr,                TEXT("CP210x.CP210xDevices")                    },
    { TEXT("CP210x.CP210xDevices\\CLSID"),                                      nullptr,                TEXT("{2FC05A0F-CD36-4A77-A788-C9BC232E947F}")  },
    { nullptr, nullptr, nullptr }
};

STDAPI DllUnregisterServer()
{
    intptr_t    x;
    HRESULT     hResult = S_OK;
    HKEY        hKey;

    if(FAILED(UnRegisterTypeLib(LIBID_CP210xExLib, 1, 0, LANG_NEUTRAL, SYS_WIN32)))
        hResult = SELFREG_E_TYPELIB;

    for(x = 0; RegEntryTable[x][0]; x++)
        ;

    while(--x >= 0)
    {
        if(RegEntryTable[x][1])
        {
            if(ERROR_SUCCESS != RegCreateKeyEx(HKEY_CLASSES_ROOT, RegEntryTable[x][0], 0, nullptr, REG_OPTION_NON_VOLATILE, KEY_WRITE, nullptr, &hKey, nullptr))
            {
                hResult = S_FALSE;
                continue;
            }

            if(ERROR_SUCCESS != RegDeleteValue(hKey, RegEntryTable[x][1]))
                hResult = S_FALSE;

            RegCloseKey(hKey);
        }
        else
        {
            if(ERROR_SUCCESS != RegDeleteKey(HKEY_CLASSES_ROOT, RegEntryTable[x][0]))
                hResult = S_FALSE;
        }
    }

    return(hResult);
}

STDAPI DllRegisterServer()
{
    ITypeLib* ptlb;
    HRESULT   hr = S_OK;
    LONG      lError;
    TCHAR     szFileName[MAX_PATH * 2];
    BOOL      bIsTypeLibReg = FALSE;
    HKEY      hKey;
    UINT      x = 0;

#ifndef UNICODE
    WCHAR  szFileNameW[MAX_PATH * 2];
#endif

    while(RegEntryTable[x][0])
    {
        if(ERROR_SUCCESS == (lError = RegCreateKeyEx(HKEY_CLASSES_ROOT, RegEntryTable[x][0], 0, nullptr, REG_OPTION_NON_VOLATILE, KEY_WRITE, nullptr, &hKey, nullptr)))
        {
            if(ENTRY_FILENAME == RegEntryTable[x][2])
            {
                GetModuleFileName(g_hInst, szFileName, _countof(szFileName));

                if(!bIsTypeLibReg)
                {
#ifndef UNICODE
                    MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED | MB_ERR_INVALID_CHARS, szFileName, -1, szFileNameW, _countof(szFileNameW));
                    if(SUCCEEDED(hr = LoadTypeLibEx(szFileNameW, REGKIND_REGISTER, &ptlb)))
#else
                    if(SUCCEEDED(hr = LoadTypeLibEx(szFileName, REGKIND_REGISTER, &ptlb)))
#endif
                    {
                        bIsTypeLibReg = TRUE;
                        ptlb->Release();
                    }
                }

                lError = RegSetValueEx(hKey, RegEntryTable[x][1], 0, REG_SZ, reinterpret_cast<BYTE*>(szFileName), static_cast<DWORD>(_tcslen(szFileName) + 1U) * sizeof(TCHAR));
            }
            else if(RegEntryTable[x][2])
            {
                lError = RegSetValueEx(hKey, RegEntryTable[x][1], 0, REG_SZ, reinterpret_cast<const BYTE*>(RegEntryTable[x][2]), static_cast<DWORD>(_tcslen(RegEntryTable[x][2]) + 1U) * sizeof(TCHAR));
            }
            RegCloseKey(hKey);
        }

        if((lError != ERROR_SUCCESS) || FAILED(hr))
        {
            DllUnregisterServer();
            return(SELFREG_E_CLASS);
        }
        ++x;
    }

    return(S_OK);
}
